@charset "utf-8";

#mouse-node {
    fill: none;
}
#canvas {
    stroke: #ccc;
    stroke-width: 7px;
    fill: white;
}
.resize-rect {
    fill: black;
    opacity: 0;
    stroke: none;
}
.label {
    font-family: sans-serif; 
    font-style: italic; 
    font-weight: bold; 
    font-size: 8px; 
    fill: black; 
    stroke: none;
    text-rendering: optimizelegibility;
    cursor: default;
}
.reaction-label {
    font-size: 30px;
    fill: rgb(32, 32, 120);
    text-rendering: optimizelegibility;
}
.node-label {
    font-size: 20px;
}
.gene-label {
    font-size: 18px;
    fill: rgb(32, 32, 120);
    text-rendering: optimizelegibility;
}
.text-label .label {
    font-size: 50px;
}
.text-label-input {
    font-size: 50px;
}
.node-circle {
    stroke-width: 2px;
}
.midmarker-circle,.multimarker-circle {
    fill: white;
    fill-opacity: 0.2;
    stroke: rgb(50, 50, 50);
}
g.selected .node-circle{
    stroke-width: 6px;
    stroke: rgb(20, 113, 199);
}
g.selected .label {
    fill: rgb(20, 113, 199);
}
.metabolite-circle {
    stroke: rgb(162, 69, 16);
    fill: rgb(224, 134, 91);
}
g.selected .metabolite-circle {
    stroke: rgb(5, 2, 0);
}
.segment {
    stroke: #334E75;
    stroke-width: 10px;
    fill: none;
}
.arrowhead {
    fill: #334E75;
}
.stoichiometry-label-rect {
    fill: white;
    opacity: 0.5;
}
.stoichiometry-label {
    fill: #334E75;
    font-size: 17px;
}
.membrane {
    fill: none;
    stroke: rgb(255, 187, 0);
}
.brush .extent { 
    fill-opacity: 0.1;
    fill: black;
    stroke: #fff;
    shape-rendering: crispEdges;
}
#brush-container .background {
    fill: none;
}
.bezier-circle {
    fill: rgb(255,255,255);     
}
.bezier-circle.b1 {
    stroke: red;
}
.bezier-circle.b2 {
    stroke: blue;
}
.connect-line{ 
    stroke: rgb(200,200,200);
}
.direction-arrow {
    stroke: black;
    stroke-width: 1px;
    fill: white;
    opacity: 0.3;
}
.start-reaction-cursor {
    cursor: pointer;
}
.start-reaction-target {
    stroke: rgb(100,100,100);
    fill: none;
    opacity: 0.5;
}
.rotation-center-line {
    stroke: red;
    stroke-width: 5px;
}
.highlight {
    fill: #D97000;
    text-decoration: underline;
}
.cursor-grab {
    cursor: grab;
    cursor: -webkit-grab;
}
.cursor-grabbing {
    cursor: grabbing;
    cursor: -webkit-grabbing;
}
.edit-text-cursor {
    cursor: text;
}
